﻿using System;
using System.Collections.Generic;
using System.Threading;

using Mcs.CMOS.Base;
using Mcs.DataTypes;
using Mcs.CMOS.StreamIO;


namespace Mcs.CMOS.DataClient
{
    public class ClientTool : ToolOutQueueBase<ClientPacket>
    {
        private PipeClient pipeClient;
        private CancellationTokenSource cTS = new CancellationTokenSource();

        public UIDataClient UserInterface { get; private set; }

        public class SpikeDataEventArgs
        {
            public SpikeDataEventArgs(Guid stream, long timeStamp, Dictionary<int, List<Spike>> data)
            {
                Stream = stream;
                TimeStamp = timeStamp;
                Data = data;
            }
            public Guid Stream { get; private set; }
            public long TimeStamp { get; private set; }
            public Dictionary<int, List<Spike>> Data { get; private set; }
        }
        public event Action<SpikeDataEventArgs> EvtSpikeData;

        public class RemoteEventEventArgs
        {
            public RemoteEventEventArgs(Event evt)
            {
                Event = evt;
            }
            public Event Event { get; private set; }
        }
        public event Action<RemoteEventEventArgs> EvtRemoteEvent;

        public enum enState
        {
            Connected,
            Disconnected
        }
        public class StateEventArgs
        {
            public StateEventArgs(enState state)
            {
                State = state;
            }
            public enState State { get; private set; }
        }
        public event Action<StateEventArgs> EvtState;



        public ClientTool(int id, string name) : 
            base(id, name, Guid.NewGuid())
        {
            UserInterface = new UIDataClient();
            UserInterface.EvtControl += OnUserInterfaceEvtControl;

            pipeClient = new PipeClient();
            pipeClient.EvtDataSpikes += OnPipeClientEvtDataSpike;
            pipeClient.EvtDataEvent += OnPipeClientEvtDataEvent;
            pipeClient.EvtState += OnPipeClientEvtState;
            pipeClient.EvtEvent += OnPipeClientEvtEvent;

            Start();
        }


        public void Close()
        {
            Break();
            pipeClient.Close();
        }

        /// <summary>
        /// Handle PipeClient events
        /// </summary>
        /// <param name="obj"></param>
        private void OnPipeClientEvtEvent(PipeClient.EventEventArgs obj)
        {
            UserInterface.LogClientEvent(obj.Event, obj.Message);

            if (obj.Event == PipeClient.enEvent.BrockenPipe)
            {
                cTS.Cancel();
            }
            else if (obj.Event == PipeClient.enEvent.ConnectionTimeOut)
            {
            }
            else if (obj.Event == PipeClient.enEvent.TaskCanceled)
            {
            }
            else if (obj.Event == PipeClient.enEvent.TransferFailed)
            {
            }
            else if (obj.Event == PipeClient.enEvent.UnhandledException)
            {
            }
        }

        /// <summary>
        /// PipeClient spike data event
        /// </summary>
        /// <remarks>
        /// Takes packet and adds it to the internal queue.
        /// </remarks>
        /// <param name="obj">Event arguments.</param>
        private void OnPipeClientEvtDataSpike(PipeClient.DataSpikesEventArgs obj)
        {            
            AddToQueue(new ClientPacket(obj.Spikes));
        }

        /// <summary>
        /// PipeClient event data event
        /// </summary>
        /// <remarks>
        /// Takes packet and adds it to the internal queue.
        /// </remarks>
        /// <param name="obj">Event arguments.</param>
        private void OnPipeClientEvtDataEvent(PipeClient.DataEventEventArgs obj)
        {
            AddToQueue(new ClientPacket(obj.Event));
        }

        /// <summary>
        /// Handle PipeClient state
        /// </summary>
        /// <param name="obj"></param>
        private void OnPipeClientEvtState(PipeClient.StateEventArgs obj)
        {
            UserInterface.SetClientState(obj.State, obj.Message);

            if(obj.State == PipeClient.enState.Connected)
            {

            }
            else if (obj.State == PipeClient.enState.Disconnected)
            {

            }
        }

        /// <summary>
        /// Handle UI commands
        /// </summary>
        /// <remarks>
        /// 
        /// </remarks>
        /// <param name="obj">Event arguments.</param>
        private async void OnUserInterfaceEvtControl(UIDataClient.ControlEventArgs obj)
        {
            if(obj.Command == UIDataClient.enCommand.ConnectDisconnect)
            {
                if(pipeClient.State == PipeClient.enState.Connected)
                {
                    cTS.Cancel();
                }
                else
                {
                    // Start client with machine name ('.' for same computer or name fo the computer in the network.)
                    pipeClient.Start(obj.Message);

                    // wait for a connection
                    cTS = new CancellationTokenSource();
                    await pipeClient.ConnectAsync(cTS.Token);

                    // Handshake with the server
                    if (pipeClient.State == PipeClient.enState.Connected)
                    {
                        if(EvtState != null)
                        {
                            EvtState(new StateEventArgs(enState.Connected));
                        }
                        // Listen to the server
                        // When a packet arrives the pipeClient sends an event
                        // OnPipeClientEvtDataSpike or OnPipeClientEvtDataEvent
                        pipeClient.ListenAsync(cTS.Token);
                    }
                }
            }
        }

        /// <summary>
        /// Handle data withdrown from the queue.
        /// </summary>
        /// <remarks>
        /// The queue is implementet in the base class. Packets are added to the queue 
        /// with the 'Work(Packet)' method of the base class. A worker thread takes 
        /// packets from the queue and evokes 'HandleData(Packet)'.
        /// </remarks>
        /// <param name="data">Data packet containing either spikes or events.</param>
        protected override void HandleData(ClientPacket data)
        {
            if (data.Content == ClientPacket.enContent.Spikes)
            {
                if (EvtSpikeData != null)
                {
                    EvtSpikeData(new SpikeDataEventArgs(Guid.Empty, 0, data.Spikes));
                }
            }
            else if (data.Content == ClientPacket.enContent.Event)
            {
                Event evt = data.Event;
                if (EvtRemoteEvent != null)
                {
                    EvtRemoteEvent(new RemoteEventEventArgs(evt));
                }

                // show all events but the timer ticks.
                if (evt.Source != Event.enSource.Timer)
                {
                    UserInterface.LogClientEvent(PipeClient.enEvent.ReceivedData, "Event: " + evt.Label + " TS = " + evt.TimeStamp.ToString());
                }
            }
        }

        public override void Initialize()
        {

        }

    }
}
